package week06

  /**
   * Main task:
   * Írjunk egy keresőfa alapú MapIntT-et, ami ebben a videóban van körülírva:
   * https://www.youtube.com/watch?v=kbPKZ1N6gSM
   *
   * A részfeladatok a trait belsejében vannak felsorolva, implementálni legtöbbjüket a traitet extendelő
   * két classban (EmptyMap és NonemptyMap) kell, de van, amit közülük lehet a traiten belül is.
    */
object MapIntT {
  def apply[T]() = EmptyMap[T]()
}

trait MapIntT[T]{
    /**
     * Implementáljunk egy
     * contains(key: Int): Boolean
     * metódust, mely visszaadja, hogy a megadott kulcs szerepel-e az MapIntTben.
     */
    def contains(key: Int): Boolean

    /**
     * Implementáljunk egy
     * size: Int
     * metódust, mely visszaadja, hogy hány kulcs-érték pár (entry) szerepel az MapIntTben.
     */
    def size: Int

    /**
     * Implementáljunk egy
     * depth: Int
     * metódust, mely visszaadja, hogy milyen mély az MapIntTet táróló fa.
     * Az üres fa magassága 0.
     */
    def depth: Int

    /**
     * Implementáljunk egy
     * isEmpty: Boolean
     * metódust, mely visszaadja, hogy üres-e az MapIntT.
     */
    def isEmpty: Boolean

    /**
     * Implementáljunk egy
     * get(key: Int): Option[T]
     * metódust, mely Some(value)-t adja vissza, ha a key kulcs létezik az MapIntTben és hozzá a value érték tartozik,
     * és None-t, ha a kulcs nem létezik az MapIntTben!
      */
    def get(key: Int): Option[T]

    /**
     * Implementáljunk egy
     * getOrElse(key: Int, default: =>T): T
     * metódust, mely a value -t adja vissza, ha a key kulcs létezik az MapIntTben és hozzá a value érték tartozik,
     * és a default értéket, ha a key kulcs nem létezik az MapIntTben!
     */
    def getOrElse(key: Int, default: =>T): T = get(key).getOrElse(default)

    /**
     * Implementáljunk egy
     * apply(key: Int): T
     * metódust, mely a value-t adja vissza, ha a key kulcs létezik az MapIntTben és hozzá a value érték tartozik,
     * és NoSuchElementException-t dob, ha a kulcs nem szerepel az MapIntTben! (ez a rész már készen van.)
     */
    def apply(key: Int): T

    /**
     * Implementáljunk egy
     * minKey: Int
     * metódust, mely az MapIntTben szereplő  minimális kulcsot adja vissza, ha van ilyen kulcs,
     * és UnsupportedOperationException-t dob, ha üres MapIntTen hívjuk! (ez a rész már készen van.)
     */
    def minKey: Int

    /**
     * Implementáljunk egy
     * maxKey: Int
     * metódust, mely az MapIntTben szereplő maximális kulcsot adja vissza, ha van ilyen kulcs,
     * és UnsupportedOperationException-t dob, ha üres MapIntTen hívjuk!
     */
    def maxKey: Int

    /**
     * Implementáljunk egy
     * remove(key: Int): MapIntT[T]
     * metódust, mely a key kulcsot a hozzá társított értékkel együtt kiveszi az MapIntTból, ha benne van,
     * és visszaadja az eredményként kapott új MapIntTet!
     * A törléshez hasznos lehet a videó.
     */
    def remove(key: Int): MapIntT[T]

    /**
     * Implementáljunk egy
     * update(key: Int, value: T): MapIntT[T]
     * metódust, mely a key kulcshoz beállítja a value értéket az MapIntT-ben és visszaadja a módosított
     * MapIntT-et! Figyeljünk arra, hogy ha eredetileg tartozott érték a kulcshoz, azt helyesen kezeljük
     * (törölni is lehet előtte, de meg lehet másképp is oldani.)
     */
    def updated(key: Int, value: T): MapIntT[T]

    /**
     * Implementáljunk egy
     * foreach[U](f: (Int,T)=>U): Unit
     * metódust, mely a mapban levő entryket **kulcs szerint növekvőbe rendezve** bejárja és mindegyik entryn
     * kiértékeli az f(entry) kifejezést!
     */
    def foreach[U](f: ((Int,T))=>U): Unit

    /**
     * Implementáljunk egy
     * map[U](f: (Int,T)=>U): Vector[U]
     * metódust, mely az MapIntTben lévő entryk mindegyikére kiértékeli **kulcs szerint növekvő sorrendben**
     * az f(entry) kifejezést, és a kapott értékeket egy vektorba pakolja!
     */
    def map[U](f: ((Int,T))=>U): Vector[U]

    /**
     * Implementáljunk egy
     * foldLeft[U](init: U)(f: (U,(Int,T))=>U): U
     * metódust, mely az MapIntT entryjein **kulcs szerint növekvő sorrendben** végighajt egy fold műveletet,
     * annak a szokásos szemantikája szerint!
     */
    def foldLeft[U](init: U)(f: (U,(Int,T))=>U): U

    /**
     * Implementáljunk egy
     * ++(m: MapIntT[T]): MapIntT[T]
     * metódust, mely a két MapIntTet "összeuniózza" úgy, hogy tkp beszúrja az m map (kulcs-érték) entryjeit
     * ebbe a mapbe és visszaadja a módosított MapIntTet! (Note: ha egy kulcs mindkét mapben szerepel, akkor
     * így az m-beli érték lesz az eredmény mapben.)
     */
    def ++(m: MapIntT[T]): MapIntT[T]

    /**
     * Implementáljunk egy
     * filter(p: (Int,T)=>Boolean): MapIntT[T]
     * metódust, mely az MapIntT-ben azokat az entryket tartja meg (és állít elő belőlük egy új MapIntT-et),
     * melyekre a p predikátum igaz!
     */
    def filter(p: ((Int,T))=>Boolean): MapIntT[T]
  }
  /**
   * A lentebbi osztályokon belül valósítsuk meg a traiten belül feladatként megadott függvényeket.
   * Amit lehet a traiten belül, azt persze lehet, hogy ott szebb lenne.
   */
  case class EmptyMap[T]() extends MapIntT[T] {
    override def contains(key: Int) = false
    override def size = 0
    override def depth = 0
    override def isEmpty = true
    override def get(key: Int) = None
    override def remove(key: Int) = this
    override def minKey = throw new UnsupportedOperationException("EmptyMap.minKey")
    override def maxKey = throw new UnsupportedOperationException("EmptyMap.maxKey")
    override def apply(key: Int) = throw new NoSuchElementException("EmptyMap.get")
    override def updated(key: Int, value: T) = NonemptyMap(this, key, value, this)
    override def foreach[U](f: ((Int,T))=>U) = ()
    override def map[U](f: ((Int,T))=>U) = Vector[U]()
    override def foldLeft[U](init: U)(f: (U, (Int, T)) => U): U = init
    override def filter(p: ((Int, T)) => Boolean): MapIntT[T] = this
    override def ++(m: MapIntT[T]): MapIntT[T] = m
  }
  case class NonemptyMap[T](left: MapIntT[T], key: Int, value: T, right: MapIntT[T]) extends MapIntT[T] {
    override def contains(key: Int) =
      if (key==this.key) true
      else if(key<this.key) left.contains(key)
      else right.contains(key)
    override def size = left.size + 1 + right.size //TODO val
    override def depth = Math.max(left.depth, right.depth) + 1 //TODO val
    override def isEmpty = false
    override def get(key: Int) =
      if (key==this.key) Some(value)
      else if(key<this.key) left.get(key)
      else right.get(key)
    override def minKey =
      if (left.isEmpty) key else left.minKey
    override def maxKey =
      if (right.isEmpty) key else right.maxKey
    override def apply(key: Int) = get(key).get
    override def remove(key: Int) =
      if (key==this.key) (left,right) match {
        case (EmptyMap(), _) => right
        case (_, EmptyMap()) => left
        case _ => {
          val nextKey = right.minKey
          val nextValue = right(nextKey)
          NonemptyMap(left, nextKey, nextValue, right.remove(nextKey))
        }
      }else if (key<this.key) NonemptyMap(left.remove(key), this.key, this.value, right)
      else NonemptyMap(left, this.key, this.value, right.remove(key))
    override def updated(key: Int, value: T) =
      if (key==this.key) NonemptyMap(left, key, value, right)
      else if(key<this.key) NonemptyMap(left.updated(key,value), this.key, this.value, right)
      else NonemptyMap(left, this.key, this.value, right.updated(key, value))
    override def foreach[U](f: ((Int,T))=>U): Unit = {
      left foreach f
      f(key,value)
      right foreach f
    }
    override def map[U](f: ((Int,T))=>U) = {
      ((left map f) :+ f(key,value)) ++ (right map f) //TODO tailrec?
    }
    override def foldLeft[U](init: U)(f: (U, (Int, T)) => U): U = {
      val leftU = (left.foldLeft(init)(f))
      val midU = f(leftU,(key,value))
      right.foldLeft(midU)(f)
    }
    override def filter(p: ((Int, T)) => Boolean): MapIntT[T] =
      if(p(key,value)) NonemptyMap( left filter p, key, value, right filter p)
      else (left filter p) ++ (right filter p)
    override def ++(m: MapIntT[T]): MapIntT[T] = //TODO
      m.foldLeft[MapIntT[T]](this) { (acc,entry) => acc.updated(entry._1, entry._2) }
  }

object Practice extends App {
  val ures = MapIntT[String]()
  println(ures)
  println( ures.updated(0,"sanyi") )
}
